# -*- coding: utf-8 -*-

__author__ = 'Daniel Donovan'
__email__ = 'spitfiredd@gmail.com'
__version__ = '0.1.1'

from collections import OrderedDict
import xmltodict
import requests
import json
import warnings
import urllib
import xml
from lxml import etree
from io import StringIO, BytesIO


warnings.filterwarnings('ignore')

field_map = {

    'piid': 'PIID',
    'idv_piid': 'REF_IDV_PIID',
    'idv_agency_id': 'REF_IDV_AGENCY_ID',
    'modification_number': 'MODIFICATION_NUMBER',

    'contracting_agency_id': 'CONTRACTING_AGENCY_ID',
    'contracting_agency_name': 'CONTRACTING_AGENCY_NAME',
    'contracting_office_id': 'CONTRACTING_OFFICE_ID',
    'contracting_office_name': 'CONTRACTING_OFFICE_NAME',
    'funding_agency_id': 'FUNDING_AGENCY_ID',
    'funding_office_id': 'FUNDING_OFFICE_ID',
    'funding_office_name': 'FUNDING_OFFICE_NAME',
    'agency_code': 'AGENCY_CODE',
    'agency_name': 'AGENCY_NAME',
    'department_id': 'DEPARTMENT_ID',
    'department_name': 'DEPARTMENT_NAME',
    'department_full_name': 'DEPARTMENT_FULL_NAME',

    'last_modified_date': 'LAST_MOD_DATE',
    'last_modified_by': 'LAST_MODIFIED_BY',
    'award_completion_date': 'AWARD_COMPLETION_DATE',
    'created_on': 'CREATED_DATE',
    'date_signed': 'SIGNED_DATE',
    'effective_date': 'EFFECTIVE_DATE',
    'estimated_completion_date': 'ESTIMATED_COMPLETION_DATE',

    'obligated_amount': 'OBLIGATED_AMOUNT',
    'ultimate_contract_value': 'ULTIMATE_CONTRACT_VALUE',
    'contract_pricing_type': 'TYPE_OF_CONTRACT_PRICING',

    'award_status': 'AWARD_STATUS',
    'contract_type': 'CONTRACT_TYPE',
    'created_by': 'CREATED_BY',
    'description': 'DESCRIPTION_OF_REQUIREMENT',
    'modification_reason': 'REASON_FOR_MODIFICATION',
    'legislative_mandates': 'LEGISLATIVE_MANDATES',
    'local_area_setaside': 'LOCAL_AREA_SET_ASIDE',
    'socioeconomic_indicators': 'SOCIO_ECONOMIC_INDICATORS',
    'multiyear_contract': 'MULTIYEAR_CONTRACT',
    'national_interest_code': 'NATIONAL_INTEREST_CODE',
    'national_interest_description': 'NATIONAL_INTEREST_DESCRIPTION',

    'naics_code': 'PRINCIPAL_NAICS_CODE',
    'naics_description': 'NAICS_DESCRIPTION',
    'product_or_service_code': 'PRODUCT_OR_SERVICE_CODE',
    'product_or_service_description': 'PRODUCT_OR_SERVICE_DESCRIPTION',

    'place_of_performance_district': 'POP_CONGRESS_DISTRICT_CODE',
    'place_of_performance_country': 'POP_CONGRESS_COUNTRY',
    'place_of_performance_state': 'POP_STATE_NAME',

    'vendor_city': 'VENDOR_ADDRESS_CITY',
    'vendor_district': 'VENDOR_CONGRESS_DISTRICT_CODE',
    'vendor_country_code': 'VENDOR_ADDRESS_COUNTRY_CODE',
    'vendor_country_name': 'VENDOR_ADDRESS_COUNTRY_NAME',
    'vendor_duns': 'VENDOR_DUNS_NUMBER',
    'vendor_dba_name': 'VENDOR_DOING_BUSINESS_AS_NAME',
    'vendor_name': 'VENDOR_NAME',
    'vendor_state_code': 'VENDOR_ADDRESS_STATE_CODE',
    'vendor_state_name': 'VENDOR_ADDRESS_STATE_NAME',
    'vendor_zip': 'VENDOR_ADDRESS_ZIP_CODE',

}

boolean_map = {
    True: 'Y',
    False: 'N',
}


html_escape_table = {
    ":": "%3A",
    "[": "%5B",
    "]": "%5D",
    ",": "%2C",
    " ": "%20",
    '"': "%22",
    '/': "%2F",
    }


class Contracts():

    feed_url = "https://www.fpds.gov/ezsearch/FEEDS/ATOM?FEEDNAME=PUBLIC&q="
    feed_size = 10
    query_url = ''

    def __init__(self, logger=None):
        #point logger to a log function, print by default
        if logger:
            self.log = logger
        else:
            self.log = print

    def pretty_print(self, data):
        self.log(json.dumps(data, indent=4))

    def html_escape(self, text):
        return "".join(html_escape_table.get(c, c) for c in text)

    def convert_params(self, params):

        new_params = {}
        for k, v in params.items():
            new_params[field_map[k]] = v
        return new_params

    def combine_params(self, params):
        if len(params) == 1:
            return " ".join("{}:{}".format(k, v) for k, v in params.items())
        else:
            return "+".join("{}:{}".format(k, v) for k, v in params.items())

    def process_data(self, data):
        # todo
        if isinstance(data, dict):
            # make a list so it's consistent
            data = [data, ]
        return data

    def get(self, num_records=100, order='desc', **kwargs):

        client = requests.session()
        # params = self.urllib.parse.quote(self.combine_params(self.convert_params(kwargs)))
        params = urllib.parse.quote(self.combine_params(self.convert_params(kwargs)),
                                    safe='+', encoding=None, errors=None)
        namespaces = {
            'http://www.fpdsng.com/FPDS': None,
            'http://www.w3.org/2005/Atom': None,
        }
        data = []
        i = 0
        # for n in range(0, num_records, 10):
        while num_records == "all" or i < num_records:

            self.log("querying {0}{1}&start={2}".format(self.feed_url,
                                                        params,
                                                        i))
            f = self.feed_url + params + '&start={0}'.format(i)
            resp = client.get(f, timeout=60, verify=False)
            self.query_url = resp.url
            self.log("finished querying {0}".format(resp.url))
            resp_data = xmltodict.parse(resp.text,
                                        process_namespaces=True,
                                        namespaces=namespaces)
            try:
                processed_data = self.process_data(resp_data['feed']['entry'])
                for pd in processed_data:
                    data.append(pd)
                    i += 1

                # if data contains less than 10 records, break out of loop
                if len(processed_data) < 10:
                    break

            except KeyError as e:
                # no results
                self.log("No results for query")
                break
        return data



# -*- coding: utf-8 -*-

__author__ = 'Daniel Donovan'
__email__ = 'spitfiredd@gmail.com'
__version__ = '0.0.1'

from collections import OrderedDict
import requests
import json
import warnings

warnings.filterwarnings('ignore')

# sample pull
# base: https://api.data.gov/sam/v4/registrations/
# duns: 0261572350000
# api: ?api_key=your-api-key

# Usage
# c = Sam('your-api-key')

# t = c.get('6928857')
# c.pretty_print(t)


class SAMImporter():

    feed_url = 'https://api.data.gov/sam/v4/registrations/'
    api = '?api_key='
    query_url = ''

    def __init__(self, api_key, logger=None):
        self.api_key = str(api_key)
        # point logger to a log function, print by default
        if logger:
            self.log = logger
        else:
            self.log = print

    def pretty_print(self, data):
        self.log(json.dumps(data, indent=4))

    def pad_contract(self, contract):
        '''
        Given a contract, this will pad leading and trailing,
        zeros. A dun number is 9 and the dun plus 4 is 13.
        Thirteen is the length of the digit passed to api.data.gov.
        '''
        return '{:0<13}'.format('{:0>9}'.format(contract))

    def get(self, duns_number):

        params = self.pad_contract(duns_number)
        query = self.feed_url + params + self.api + self.api_key
        self.log("querying {}".format(query))
        resp = requests.get(query, verify=False, timeout=60)
        self.query_url = resp.url
        self.log("finished querying {}".format(self.query_url))
        try:
            data = json.loads(resp.text)
        except ValueError:
            self.log("No results for query")

        return data



'''
File of dictionary, list, and pandas utility functions
'''
from functools import reduce
import csv
from datetime import date
from dateutil.relativedelta import relativedelta
import pandas as pd

# from g2x_helpers.pyfpds import Contracts


def nested_get(_dict, keys, default=None, _sep='.'):
    ''' Returns the value from a dict with deep nested keys
    '''
    if isinstance(keys, str):
        _keys = keys.split(_sep)
    else:
        _keys = keys

    def _reducer(d, key):
        if isinstance(d, dict):
            return d.get(key, default)
        return default
    return reduce(_reducer, _keys, _dict)


def flatten_dict(d):
    """Flatten a nested diction
    """
    def items():
        for key, value in d.items():
            if isinstance(value, dict):
                for subkey, subvalue in flatten_dict(value).items():
                    yield key + "." + subkey, subvalue
            else:
                yield key, value
    return dict(items())


def create_mapping(filename):
    """
    Creates a python dictionay from a csv file. The csv must only hold
    two columns of data.
    """
    reader = csv.reader(open(str(filename), 'r'))
    _d = {}
    for row in reader:
        k, v = row
        _d[k] = v
    return _d


def date_between_list(start=None, stop=date.today()):
    date_list = []
    # stop = date.today()
    start = start
    end = start + relativedelta(day=1, months=12, days=-1)
    while True:
        if end > stop:
            dates_between = "[{:%Y/%m/%d},{:%Y/%m/%d}]".format(start, stop)
            date_list.append(dates_between)
            break
        else:
            dates_between = "[{:%Y/%m/%d},{:%Y/%m/%d}]".format(start, end)
            start = end + relativedelta(day=+1, months=1)
            end = start + relativedelta(day=1, months=12, days=-1)
            date_list.append(dates_between)
    return date_list


def convert_to_null(dataframe, *cols):
    dataframe['referenced_idvid_piid'] = dataframe['referenced_idvid_piid'].fillna(0)
    out = dataframe.where((pd.notnull(dataframe)), None)
    return out


def create_dict(dataframe):
    return dataframe.to_dict(orient='records')


def pull_data(**kwargs):
    c = Contracts()
    records = c.get(**kwargs)
    return records


def create_keep_list(filename):
    df = pd.read_csv(filename)
    return df['keepList'].tolist()


def apply_keep_list_to_frame(df, keep_list):
    return df[keep_list]


def check_if_col_exists(dataframe, list_):
    for item in list_:
        if item not in dataframe:
            dataframe[item] = 0
    return dataframe


def data_pull(naics_list=None, dept=None, idv_mapping=None,
              award_mapping=None, keep_list=None, **kwargs):
    combined_df = pd.DataFrame()
    d_name = '"{}"'.format(dept)
    for naics in naics_list:
        c = Contracts()
        records = c.get(naics_code=naics, department_name=d_name, **kwargs)
        for record in records:
            if 'IDV' in record['content']:
                _ivdd = flatten_dict(record)
                _idvl = list()
                _idvl.append(_ivdd)
                _tmp = pd.DataFrame(_idvl)
                _tmp = _tmp.rename(index=str, columns=idv_mapping)
                _tmp = apply_keep_list_to_frame(_tmp, keep_list)
                # idv = idv.append(_tmp)
                combined_df = combined_df.append(_tmp)
            elif 'award' in record['content']:
                _awardsd = flatten_dict(record)
                _awardsl = list()
                _awardsl.append(_awardsd)
                _tmp1 = pd.DataFrame(_awardsl)
                _tmp1 = _tmp1.rename(index=str, columns=award_mapping)
                _tmp1 = apply_keep_list_to_frame(_tmp1, keep_list)
                # award = award.append(_tmp1)
                combined_df = combined_df.append(_tmp1)
    combined_df = convert_to_null(combined_df,
                                  'referenced_idvid_piid')
    return combined_df


def data_pull_funding_agency(idv_mapping=None,
                             award_mapping=None,
                             award_keep=None,
                             ivd_keep=None,
                             naics_list=None,
                             funding_agency=None,
                             **kwargs):
    combined_df = pd.DataFrame()
    d_name = '"{}"'.format(funding_agency)
    for naics in naics_list:
        c = Contracts()
        records = c.get(naics_code=naics, num_records="all",
                        funding_agency_id=d_name, **kwargs)
        for record in records:
            if 'IDV' in record['content']:
                _ivdd = flatten_dict(record)
                _idvl = list()
                _idvl.append(_ivdd)
                _tmp = pd.DataFrame(_idvl)
                _tmp = _tmp.rename(index=str, columns=idv_mapping)
                _tmp = check_if_col_exists(_tmp, award_keep)
                _tmp = apply_keep_list_to_frame(_tmp, award_keep)
                # idv = idv.append(_tmp)
                combined_df = combined_df.append(_tmp)
            elif 'award' in record['content']:
                _awardsd = flatten_dict(record)
                _awardsl = list()
                _awardsl.append(_awardsd)
                _tmp1 = pd.DataFrame(_awardsl)
                _tmp1 = _tmp1.rename(index=str, columns=award_mapping)
                _tmp1 = check_if_col_exists(_tmp1, ivd_keep)
                _tmp1 = apply_keep_list_to_frame(_tmp1, ivd_keep)
                # award = award.append(_tmp1)
                combined_df = combined_df.append(_tmp1)
    combined_df = convert_to_null(combined_df,
                                  'referenced_idvid_piid')
    return combined_df


def data_pull_general(idv_mapping=None,
                      award_mapping=None,
                      award_keep=None,
                      ivd_keep=None,
                      naics_list=None,
                      **kwargs):
    combined_df = pd.DataFrame()
    for naics in naics_list:
        c = Contracts()
        records = c.get(naics_code=naics, num_records="all", **kwargs)
        for record in records:
            if 'IDV' in record['content']:
                _ivdd = flatten_dict(record)
                _idvl = list()
                _idvl.append(_ivdd)
                _tmp = pd.DataFrame(_idvl)
                _tmp = _tmp.rename(index=str, columns=idv_mapping)
                _tmp = check_if_col_exists(_tmp, award_keep)
                _tmp = apply_keep_list_to_frame(_tmp, award_keep)
                # idv = idv.append(_tmp)
                combined_df = combined_df.append(_tmp)
            elif 'award' in record['content']:
                _awardsd = flatten_dict(record)
                _awardsl = list()
                _awardsl.append(_awardsd)
                _tmp1 = pd.DataFrame(_awardsl)
                _tmp1 = _tmp1.rename(index=str, columns=award_mapping)
                _tmp1 = check_if_col_exists(_tmp1, ivd_keep)
                _tmp1 = apply_keep_list_to_frame(_tmp1, ivd_keep)
                # award = award.append(_tmp1)
                combined_df = combined_df.append(_tmp1)
    combined_df = convert_to_null(combined_df,
                                  'referenced_idvid_piid')
    return combined_df



from setuptools import setup

setup(name='g2x_helpers',
      version='0.4',
      description='Helper Functions used across different packages',
      author='Daniel Donovan',
      author_email='spitfiredd@gmail.com',
      url='https://github.com/spitfiredd/g2x_helpers',
      license='MIT',
      packages=['g2x_helpers', 'g2x_helpers.pysam', 'g2x_helpers.pyfpds'],
      install_requires=['requests', 'xmltodict'],
      zip_safe=False)
